# SPDX-License-Identifier: Apache-2.0

from datetime import datetime, timezone, timedelta
from base64 import urlsafe_b64encode
from os import urandom
import mimetypes

from argon2 import PasswordHasher
from itsdangerous.url_safe import URLSafeTimedSerializer
from sqlalchemy import ForeignKey, Text, Boolean, DateTime, func, create_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from foxfiles.settings import SECRET_KEY, DATABASE, ECHO_QUERIES

signer = URLSafeTimedSerializer(SECRET_KEY)
ph = PasswordHasher()


class Base(DeclarativeBase):
    pass


class User(Base):
    __tablename__ = "users"

    id: Mapped[int] = mapped_column(primary_key=True)
    username: Mapped[str] = mapped_column(Text(), unique=True)
    password: Mapped[str] = mapped_column(Text())
    is_admin: Mapped[bool] = mapped_column(Boolean(), default=False)
    password_changed_at: Mapped[datetime] = mapped_column(
        DateTime(), default=func.now()
    )
    created_at: Mapped[datetime] = mapped_column(DateTime(), default=func.now())
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(), default=func.now(), onupdate=func.now()
    )

    files: Mapped[list["File"]] = relationship(
        back_populates="user", cascade="all, delete-orphan"
    )

    def __init__(self, username: str, password: str, is_admin: bool = False):
        self.username = username
        self.password = ph.hash(password)
        # the value returned by the timestamp signer is only precise to the nearest second, while this is precise to the millisecond.
        # so tokens generated immediately after user creation/changing passwords will not be valid.
        # this is fine for existing users--it takes a few seconds to log back in and generate the first new token, anyway
        # but it breaks signing up
        self.password_changed_at = datetime.utcnow() - timedelta(seconds=30)
        self.is_admin = is_admin

    def __repr__(self):
        return f"User(id={self.id}, username='{self.username}')"

    def set_password(self, password: str):
        """Sets a password, but doesn't save it."""

        self.password = ph.hash(password)
        self.password_changed_at = datetime.utcnow()

    def verify_password(self, password: str):
        """Checks the user's password.
        Returns True if the password is valid, False otherwise.
        Does not raise an error."""

        try:
            return ph.verify(self.password, password)
        except:
            return False

    def get_token(self):
        return signer.dumps(str(self.id))

    def verify_token(self, token: str):
        try:
            value, timestamp = signer.loads(token, return_timestamp=True)
            first_valid = self.password_changed_at.replace(tzinfo=timezone.utc)
            return timestamp > first_valid and value == str(self.id)
        except:
            return False


def _random_url_id():
    return urlsafe_b64encode(urandom(12)).decode("utf-8")


class File(Base):
    __tablename__ = "files"

    id: Mapped[int] = mapped_column(primary_key=True)
    url_id: Mapped[str] = mapped_column(Text(), unique=True, default=_random_url_id)
    filename: Mapped[str] = mapped_column(Text())
    hash: Mapped[str] = mapped_column(Text())
    content_type: Mapped[str] = mapped_column(Text())

    created_at: Mapped[datetime] = mapped_column(DateTime(), default=func.now())
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(), default=func.now(), onupdate=func.now()
    )

    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
    user: Mapped[User] = relationship(back_populates="files")

    @property
    def path(self):
        ext = mimetypes.guess_extension(self.content_type, strict=False)
        if not ext:
            ext = ""

        return f"{self.url_id}{ext}"



engine = create_engine(f"sqlite:///{DATABASE}", echo=ECHO_QUERIES)

Base.metadata.create_all(engine)
