# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
import mimetypes
import hashlib
import os

import magic
from werkzeug.datastructures import FileStorage

from foxfiles.settings import STORAGE_BACKEND, STORAGE_LOCAL_DIR


if STORAGE_BACKEND == "local":
    path = Path(".") / STORAGE_LOCAL_DIR
    os.makedirs(path, exist_ok=True)


def _local_file_path(hash: str, content_type: str):
    ext = mimetypes.guess_extension(content_type, strict=False)
    if not ext:
        raise ValueError("Extension couldn't be guessed")

    return Path(".") / STORAGE_LOCAL_DIR / f"{hash}{ext}"


def get_file_by_hash(hash: str, content_type: str):
    match STORAGE_BACKEND:
        case "local":
            return _local_get_file_by_hash(hash, content_type)
        case "s3":
            raise NotImplementedError()


def upload_file(file: FileStorage) -> tuple[str, str]:
    """Upload a file.
    The first return argument is the hash, the second return argument is the file's MIME type.
    """

    match STORAGE_BACKEND:
        case "local":
            return _local_upload_file(file)
        case "s3":
            raise NotImplementedError()


def delete_file(hash: str, content_type: str):
    match STORAGE_BACKEND:
        case "local":
            return _local_delete_file(hash, content_type)
        case "s3":
            raise NotImplementedError()


def _local_get_file_by_hash(hash: str, content_type: str):
    p = _local_file_path(hash, content_type)
    with p.open("rb") as f:
        return f.read()


def _local_upload_file(file: FileStorage) -> tuple[str, str]:
    """Uploads `file` to local storage.
    The first return argument is the hash, the second return argument is the file's MIME type.
    """

    stream = file.stream.read()
    file_type = magic.from_buffer(stream[:2048], mime=True)
    hash = hashlib.sha256(stream).hexdigest()
    p = _local_file_path(hash, file_type)

    with p.open("wb") as f:
        f.write(stream)

    return (hash, file_type)


def _local_delete_file(hash: str, content_type: str):
    p = _local_file_path(hash, content_type)
    try:
        os.remove(p)
    except FileNotFoundError:
        pass
